%  ===================================================
%       esempio del metodo degli elementi finiti
%  ===================================================
%  Risolviamo una equazione del II ordine
%  nell?intervallo: [0,1];
%     - k*u''(x) = s(x)
%  ponendo per k il valore 3 e usando
%  il metodo  degli elementi finiti.
%  Esaminiamo un caso a soluzione nota e confrontiamo
%  i risultati ottenuti con il FEM con quelli teorici.
%  Consideriamo la funzione
%     u(x) = x*x*x;   u'(x) = 3*x*x    u''(x)= 6*x;
%  la funzione s(x) risulta s(x) = -18*x;
%  Come condizioni al contorno del problema imponiamo
%  il valore della funzione ad una estremit
%  con u(0)=0 e il valore del flusso all?estremit 
%  opposta (valore del flusso per x=1) u'(1)=3.

clear  % azzera le variabili in memoria
clc % pulisci finestra di testo

%  ===================================================
%        seleziona modalita? grafica
%  ===================================================
      
close  % chiudi finestra grafica precedente

h1 = figure(1);  %  apre una finestra grafica
set(h1, 'Units', 'normalized',...
    'Position', [0.5 0.5 0.45 0.5]);
whitebg(h1, 'black') %  colore di fondo

hold on % piu? tracciamenti sulla stessa finestra
zoom on % ingrandire l?immagine

%  ===================================================
%             grafico funzione esatta
%  ===================================================

XX = 0:0.01:1; % crea il vettore delle ascisse
N = size(XX,2); % numero colonne del vettore
for i = 1:N;
   %valori della funzione u
   UU(i) = XX(i)*XX(i)*XX(i);
end
plot(XX,UU);   % grafica la funzione

 %  ===================================================
 %             divido in intervalli (anche diseguali)
 %  ===================================================
                     
 X = [0  0.2  0.4  0.5  0.6  0.8 1]; % assegna nodi
 n = size(X,2);  % numero elementi del vettore X
 k = 3;          % costante materiale
 
 %  imposizione delle condizioni al contorno:
 %  valutiamo u(0) e la assegniamo ad U(1);
 
 U(1) = 0 % valore iniziale [U= 0]
 
 %  valutiamo u?(1)
 Duf = 3; % valore finale della derivata [Duf= 3*1*1]
 d = k*Duf;
 
 %  ===================================================
 %             calcolo dei coefficienti
 %  ===================================================
 
 %  Devo formare il sistema:
 
%      Somma su h   f(i,h) U(h) = s(i)+c(i)
% -----------------------------------------------------------
%  f(1,1)*U(1)+f(1,1)*U(2)+...   = s(1)+c(1)
%    ...................
%    ... +f(h,h-1)*U(h-1)+f(h,h)*U(h)+f(h,h+1)*U(h+1)
%                                                      +... = s(h)+c(h)
%    ...................
% ...+f(n,n-1)*U(n-1)+f(n,n)*U(n)= s(n)+c(n)
% -----------------------------------------------------------
 
%  Inizialmente si assegna zero a tutti gli elementi matrice
f=zeros(n,n);
%  la prima equazione viene sostituita con U(1)=0 quindi
f(1,1)=1;
%  pertanto l?indice h parte da 2
for h = 2:(n-1)
    f(h,h-1) = -k/(X(h)-X(h-1));
    f(h,h)   = +k/(X(h)-X(h-1)) + k/(X(h+1)-X(h));
    f(h,h+1) = -k/(X(h+1)-X(h));
end
f(n,n-1) = - k/(X(n)-X(n-1));
f(n,n)   = + k/(X(n)-X(n-1));
 
% =================================================== 
% termini di sorgente s(i) nei singoli elementi
% calcolati con due punti di Gauss
% =================================================== 
% S(1)  inutile: la prima equazione va sostituita.
%  Punti di Gauss per un polinomio di secondo e terzo grado                 
g1=-1/sqrt(3); 
g2=+1/sqrt(3);
              
%  ----------------------------------------
for h=2:(n-1)
% coordinate estremi degli intervalli adiacenti:
    A=X(h-1); B=X(h);   C=X(h+1);
% coordinate dei punti di Gauss dei due segmenti:
    x1=(A+B)/2+(B-A)/2*g1; x2=(A+B)/2+(B-A)/2*g2;
    x3=(B+C)/2+(C-B)/2*g1; x4=(B+C)/2+(C-B)/2*g2;
 
% le funzioni da integrare sono:
% da integrare su [A, B]:   F= (x-A)*(-18*x)/(B-A);
% da integrare su [B, C]:  G=(C-x)*(-18*x)/(C-B);
% Pertanto i valori degli integrali nei
% rispettivi intervalli sono:
    
    S1=((x1-A)*(-18*x1)+(x2-A)*(-18*x2))/2;
    S2=((C-x3)*(-18*x3)+(C-x4)*(-18*x4))/2;
    S(h)=S1+S2;
end

%  coordinate estremi dell?ultimo intervallo
A=X(n-1); B=X(n);
%  coordinate dei punti di Gauss dell?ultimo intervallo:
x1=(A+B)/2+(A-B)/2*g1;  x2=(A+B)/2+(A-B)/2*g2;
S(n)=((x1-A)*(-18*x1)+( x2-A)*(-18*x2))/2;

% =================================================== 
% termini di contorno C(i)
% =================================================== 
% C(1) non si pu calcolare: e del resto non serve.
for h = 2:(n-1)
    C(h) = 0;
end
C(n) = d;
%  Sommo i termini di sorgente con quelli di bordo
SC(1)=0; % valore iniziale della U
for h = 2:n
    SC(h) = S(h) + C(h) ;
end
                     
U = f \ SC';

plot(X,U,'r','era','back') % soluzione approssimata
                     
